<?php

namespace Drupal\oidc_mcpf\Form;

use Drupal\Core\Asset\LibraryDiscoveryInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form to edit the module configuration.
 */
class ConfigForm extends ConfigFormBase {

  /**
   * The user roles storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $userRoleStorage;

  /**
   * The library discovery service.
   *
   * @var \Drupal\Core\Asset\LibraryDiscoveryInterface
   */
  protected $libraryDiscovery;

  /**
   * Class constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Asset\LibraryDiscoveryInterface $library_discovery
   *   The library discovery service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, LibraryDiscoveryInterface $library_discovery) {
    parent::__construct($config_factory);

    $this->libraryDiscovery = $library_discovery;
    $this->userRoleStorage = $entity_type_manager->getStorage('user_role');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('library.discovery')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'oidc_mcpf.general',
      'oidc_mcpf.idm',
      'oidc_mcpf.toolbar',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'oidc_mcpf_config_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['general'] = [
      '#type' => 'details',
      '#title' => $this->t('General'),
      '#open' => TRUE,
      '#tree' => TRUE,
    ];

    $form['general'] += $this->buildGeneralForm($form_state);

    $form['idm'] = [
      '#type' => 'details',
      '#title' => $this->t('IDM'),
      '#open' => TRUE,
      '#tree' => TRUE,
    ];

    $form['idm'] += $this->buildIdmForm($form_state);

    $form['toolbar'] = [
      '#type' => 'details',
      '#title' => $this->t('Toolbar'),
      '#open' => TRUE,
      '#tree' => TRUE,
    ];

    $form['toolbar'] += $this->buildToolbarForm($form_state);

    return parent::buildForm($form, $form_state);
  }

  /**
   * Build the general element form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Current state of the form.
   *
   * @return array
   *   The form structure.
   */
  protected function buildGeneralForm(FormStateInterface $form_state) {
    $config = $this->config('oidc_mcpf.general');

    $element = [];

    $element['environment'] = [
      '#type' => 'radios',
      '#title' => $this->t('Environment'),
      '#options' => [
        'test' => $this->t('Test'),
        'production' => $this->t('Production'),
      ],
      '#default_value' => $config->get('environment'),
      '#required' => TRUE,
    ];

    return $element;
  }

  /**
   * Build the IDM element form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Current state of the form.
   *
   * @return array
   *   The form structure.
   */
  protected function buildIdmForm(FormStateInterface $form_state) {
    $config = $this->config('oidc_mcpf.idm');
    $roles = $this->userRoleStorage->loadMultiple();

    // Remove the locked roles.
    unset($roles[UserInterface::ANONYMOUS_ROLE], $roles[UserInterface::AUTHENTICATED_ROLE]);

    // Turn into an [ID => label] array.
    array_walk($roles, static function(&$role) {
      $role = $role->label();
    });

    // Store the roles mapping.
    if (!$form_state->has('idm_roles_mapping')) {
      $roles_mapping = $config->get('roles_mapping') ?: [];

      if ($roles_mapping) {
        $roles_mapping = array_intersect_key($roles_mapping, $roles);
      }

      $form_state->set('idm_roles_mapping', $roles_mapping);
    }

    $element = [];

    $element['roles_claim'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Roles claim'),
      '#description' => $this->t('Don\'t forget to add the matching scope to the <a href=":url">ACM configuration</a>.', [
        ':url' => Url::fromRoute('oidc.admin.realms_config')->toString(TRUE)->getGeneratedUrl(),
      ]),
      '#default_value' => $config->get('roles_claim'),
    ];

    $element['roles_mapping'] = [
      '#type' => 'details',
      '#title' => $this->t('Roles mapping'),
      '#description' => $this->t('Specify the IDM roles in the same format as configured in ACM: @formats. In case ACM provides the roles with a scope, the format with role and context is also supported.', [
        '@formats' => _oidc_format_list([
          '<role>-<context>:<scope>',
          '<role>:<context>',
          '<role>',
        ]),
      ]),
      '#open' => TRUE,
      '#parents' => ['idm', '_roles_mapping'],
      '#prefix' => '<div id="oidc-mcpf-idm-roles-mapping">',
      '#suffix' => '</div>',
      '#states' => [
        'visible' => [
          ':text[name="idm[roles_claim]"]' => ['filled' => TRUE],
        ],
      ],
    ];

    $element['roles_mapping']['table'] = [
      '#type' => 'table',
      '#header' => [
        $this->t('Drupal role'),
        $this->t('IDM role'),
        $this->t('Delete'),
      ],
      '#empty' => $this->t('No roles mapping has been specified.'),
    ];

    foreach ($form_state->get('idm_roles_mapping') as $rid => $idm_role) {
      $row = [];

      $row['drupal_role'] = [
        '#plain_text' => $roles[$rid],
      ];

      $row['idm_role'] = [
        '#type' => 'textfield',
        '#title' => $this->t('IDM role'),
        '#title_display' => 'invisible',
        '#default_value' => $idm_role,
        '#parents' => ['idm', 'roles_mapping', $rid],
      ];

      $row['delete'] = [
        '#type' => 'submit',
        '#name' => 'delete_idm_' . $rid,
        '#value' => $this->t('Delete'),
        '#submit' => ['::deleteIdmRolesMappingEntrySubmit'],
        '#limit_validation_errors' => [],
        '#ajax' => [
          'wrapper' => 'oidc-mcpf-idm-roles-mapping',
          'callback' => '::idmRolesMappingAjax',
        ],
        '#rid' => $rid,
      ];

      $element['roles_mapping']['table'][$rid] = $row;

      unset($roles[$rid]);
    }

    $element['roles_mapping']['add'] = [
      '#type' => 'details',
      '#title' => $this->t('Add role'),
      '#open' => TRUE,
      '#attributes' => [
        'class' => ['container-inline'],
      ],
      '#access' => !empty($roles),
    ];

    $element['roles_mapping']['add']['rid'] = [
      '#type' => 'select',
      '#title' => $this->t('Drupal role'),
      '#title_display' => 'invisible',
      '#options' => $roles,
      '#empty_option' => '- ' . $this->t('Select') . ' -'
    ];

    $element['roles_mapping']['add']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add'),
      '#validate' => ['::addIdmRolesMappingEntryValidate'],
      '#submit' => ['::addIdmRolesMappingEntrySubmit'],
      '#limit_validation_errors' => [
        ['idm', '_roles_mapping', 'add'],
      ],
      '#ajax' => [
        'wrapper' => 'oidc-mcpf-idm-roles-mapping',
        'callback' => '::idmRolesMappingAjax',
      ],
    ];

    return $element;
  }

  /**
   * Build the toolbar element form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Current state of the form.
   *
   * @return array
   *   The form structure.
   */
  protected function buildToolbarForm(FormStateInterface $form_state) {
    $config = $this->config('oidc_mcpf.toolbar');

    $element = [];

    $element['enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enabled'),
      '#default_value' => (int) $config->get('enabled'),
    ];

    $element['toolbar'] = [
      '#type' => 'container',
      '#parents' => ['toolbar'],
      '#states' => [
        'visible' => [
          ':checkbox[name="toolbar[enabled]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $element['toolbar']['widget_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Widget ID'),
      '#default_value' => $config->get('widget_id'),
      '#states' => [
        'required' => [
          ':checkbox[name="toolbar[enabled]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $element['toolbar']['anonymous'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Show for anonymous users'),
      '#default_value' => (int) $config->get('anonymous'),
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Validate the IDM roles.
    if ($form_state->getValue(['idm', 'roles_claim'])) {
      $roles_mapping = $form_state->getValue(['idm', 'roles_mapping'], []);

      foreach ($roles_mapping as $rid => $idm_role) {
        if ($idm_role === '') {
          $form_state->setErrorByName('idm][roles_mapping][' . $rid, $this->t('The mapped IDM role must be specified.'));
        }
        elseif (!preg_match('/^[A-Za-z]+(?::[a-z]+|-[a-z0-9]+:[A-Z0-9]+)?$/', $idm_role)) {
          $form_state->setErrorByName('idm][roles_mapping][' . $rid, $this->t('The IDM role is not in a valid format.'));
        }
      }
    }

    // Ensure a widget ID is specified.
    if ($form_state->getValue(['toolbar', 'enabled']) && !$form_state->getValue(['toolbar', 'widget_id'])) {
      $form_state->setErrorByName('toolbar][widget_id', $this->t('@title field is required.', [
        '@title' => $this->t('Widget ID'),
      ]));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // General configuration.
    $this->config('oidc_mcpf.general')
      ->set('environment', $form_state->getValue(['general', 'environment']))
      ->save();

    // IDM configuration.
    $roles_claim = $form_state->getValue(['idm', 'roles_claim']);
    $roles_mapping = [];

    if ($roles_claim) {
      $roles_mapping = $form_state->getValue(['idm', 'roles_mapping']);
    }

    $this->config('oidc_mcpf.idm')
      ->set('roles_claim', $roles_claim ?: NULL)
      ->set('roles_mapping', $roles_mapping ?: NULL)
      ->save();

    // Toolbar configuration.
    $config = $this->config('oidc_mcpf.toolbar');
    $enabled = $config->get('enabled');

    $config
      ->set('enabled', (bool) $form_state->getValue(['toolbar', 'enabled']))
      ->set('widget_id', $form_state->getValue(['toolbar', 'widget_id']) ?: NULL)
      ->set('anonymous', (bool) $form_state->getValue(['toolbar', 'anonymous']))
      ->save();

    if ($enabled || $config->get('enabled')) {
      $this->libraryDiscovery->clearCachedDefinitions();
    }

    parent::submitForm($form, $form_state);
  }

  /**
   * Form validation handler; Validates the new IDM roles mapping entry.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function addIdmRolesMappingEntryValidate(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getValue(['idm', '_roles_mapping', 'add', 'rid'])) {
      $form_state->setErrorByName('idm][_roles_mapping][add][rid', $this->t('Select the role to add.'));
    }
  }

  /**
   * Form submit handler; Add an entry to the IDM roles mapping.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function addIdmRolesMappingEntrySubmit(array &$form, FormStateInterface $form_state) {
    $roles_mapping = &$form_state->get('idm_roles_mapping');
    $rid = $form_state->getValue(['idm', '_roles_mapping', 'add', 'rid']);

    $roles_mapping[$rid] = '';

    $user_input = $form_state->getUserInput();
    unset($user_input['idm']['_roles_mapping']['add']);
    $form_state->setUserInput($user_input);

    $form_state->setRebuild();
  }

  /**
   * Form submit handler; Delete an entry from the IDM roles mapping.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function deleteIdmRolesMappingEntrySubmit(array &$form, FormStateInterface $form_state) {
    $roles_mapping = &$form_state->get('idm_roles_mapping');
    $rid = $form_state->getTriggeringElement()['#rid'];

    unset($roles_mapping[$rid]);

    $form_state->setRebuild();
  }

  /**
   * Ajax callback; Returns the ajax response when the IDM roles mapping was changed.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   Replacement content for the IDM roles mapping element.
   */
  public function idmRolesMappingAjax(array $form, FormStateInterface $form_state) {
    return $form['idm']['roles_mapping'];
  }

}
